# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
IFPS Views
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model

from tailbone.views import MasterView


class IFPS_PLUView(MasterView):
    """
    Master view for the Department class.
    """
    model_class = model.IFPS_PLU
    route_prefix = 'ifps_plus'
    url_prefix = '/ifps-plu-codes'
    results_downloadable = True
    has_versions = True

    labels = {
        'plu': "PLU",
        'gpc': "GPC",
        'aka': "AKA",
        'measurements_north_america': "Measurements (North America)",
        'measurements_rest_of_world': "Measurements (rest of world)",
    }

    grid_columns = [
        'plu',
        'category',
        'commodity',
        'variety',
        'size',
        'botanical_name',
        'revision_date',
    ]

    def configure_grid(self, g):
        super(IFPS_PLUView, self).configure_grid(g)

        g.filters['plu'].default_active = True
        g.filters['plu'].default_verb = 'equal'

        g.filters['commodity'].default_active = True
        g.filters['commodity'].default_verb = 'contains'

        g.set_sort_defaults('plu')

        # variety
        # this is actually a TEXT field, so potentially large
        g.set_renderer('variety', self.render_truncated_value)

        g.set_link('plu')
        g.set_link('commodity')
        g.set_link('variety')

    def configure_form(self, f):
        super(IFPS_PLUView, self).configure_form(f)

        if self.creating:
            f.remove('revision_date',
                     'date_added')
        else:
            f.set_readonly('revision_date')
            f.set_readonly('date_added')



def includeme(config):
    IFPS_PLUView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for tempmon probes
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six

from rattail.time import make_utc, localtime
from rattail_tempmon.db import model as tempmon

import colander
from deform import widget as dfwidget
from webhelpers2.html import tags, HTML

from tailbone import forms, grids
from tailbone.views.tempmon import MasterView
from tailbone.util import raw_datetime


class TempmonProbeView(MasterView):
    """
    Master view for tempmon probes.
    """
    model_class = tempmon.Probe
    model_title = "TempMon Probe"
    model_title_plural = "TempMon Probes"
    route_prefix = 'tempmon.probes'
    url_prefix = '/tempmon/probes'

    has_rows = True
    model_row_class = tempmon.Reading

    labels = {
        'critical_max_timeout': "Critical High Timeout",
        'good_max_timeout': "High Timeout",
        'good_min_timeout': "Low Timeout",
        'critical_min_timeout': "Critical Low Timeout",
    }

    grid_columns = [
        'client',
        'config_key',
        'appliance',
        'appliance_type',
        'description',
        'device_path',
        'enabled',
        'status',
    ]

    form_fields = [
        'client',
        'config_key',
        'appliance',
        'appliance_type',
        'description',
        'location',
        'device_path',
        'critical_temp_max',
        'critical_max_timeout',
        'good_temp_max',
        'good_max_timeout',
        'good_temp_min',
        'good_min_timeout',
        'critical_temp_min',
        'critical_min_timeout',
        'error_timeout',
        'therm_status_timeout',
        'status_alert_timeout',
        'notes',
        'enabled',
        'status',
    ]

    row_grid_columns = [
        'degrees_f',
        'taken',
    ]

    def configure_grid(self, g):
        super(TempmonProbeView, self).configure_grid(g)

        g.joiners['client'] = lambda q: q.join(tempmon.Client)
        g.sorters['client'] = g.make_sorter(tempmon.Client.config_key)
        g.set_sort_defaults('client')

        g.set_enum('appliance_type', self.enum.TEMPMON_APPLIANCE_TYPE)
        g.set_enum('status', self.enum.TEMPMON_PROBE_STATUS)

        g.set_renderer('enabled', self.render_enabled_grid)

        g.set_label('config_key', "Key")

        g.set_link('client')
        g.set_link('config_key')
        g.set_link('description')

    def render_enabled_grid(self, probe, field):
        if probe.enabled:
            return "Yes"
        return "No"

    def configure_form(self, f):
        super(TempmonProbeView, self).configure_form(f)

        # config_key
        f.set_validator('config_key', self.unique_config_key)

        # client
        f.set_renderer('client', self.render_client)
        f.set_label('client', "Tempmon Client")
        if self.creating or self.editing:
            f.replace('client', 'client_uuid')
            clients = self.Session.query(tempmon.Client)
            if self.creating:
                clients = clients.filter(tempmon.Client.archived == False)
            clients = clients.order_by(tempmon.Client.config_key)
            client_values = [(client.uuid, "{} ({})".format(client.config_key, client.hostname))
                             for client in clients]
            f.set_widget('client_uuid', dfwidget.SelectWidget(values=client_values))
            f.set_label('client_uuid', "Tempmon Client")

        # appliance
        f.set_renderer('appliance', self.render_appliance)
        if self.creating or self.editing:
            f.replace('appliance', 'appliance_uuid')
            appliances = self.Session.query(tempmon.Appliance)\
                                     .order_by(tempmon.Appliance.name)
            appliance_values = [(appliance.uuid, appliance.name)
                                for appliance in appliances]
            appliance_values.insert(0, ('', "(none)"))
            f.set_widget('appliance_uuid', dfwidget.SelectWidget(values=appliance_values))
            f.set_label('appliance_uuid', "Appliance")

        # appliance_type
        f.set_enum('appliance_type', self.enum.TEMPMON_APPLIANCE_TYPE)

        # therm_status_timeout
        f.set_helptext('therm_status_timeout', tempmon.Probe.therm_status_timeout.__doc__)

        # status_alert_timeout
        f.set_helptext('status_alert_timeout', tempmon.Probe.status_alert_timeout.__doc__)

        # notes
        f.set_type('notes', 'text')

        # status
        f.set_enum('status', self.enum.TEMPMON_PROBE_STATUS)
        if self.creating or self.editing:
            f.remove_fields('status')

        # enabled
        if self.creating or self.editing:
            f.set_node('enabled', forms.types.DateTimeBoolean())
        else:
            f.set_renderer('enabled', self.render_enabled_form)
        f.set_helptext('enabled', tempmon.Probe.enabled.__doc__)

    def objectify(self, form, data=None):

        # this is a hack to prevent updates to the 'enabled' timestamp, when
        # simple edits are being done to the probe.  i.e. we do want to set the
        # timestamp when it was previously null, but not otherwise.
        if self.editing:
            data = dict(data or form.validated)
            if data['enabled'] and form.model_instance.enabled:
                data['enabled'] = form.model_instance.enabled

        return super(TempmonProbeView, self).objectify(form, data=data)

    def unique_config_key(self, node, value):
        query = self.Session.query(tempmon.Probe)\
                            .filter(tempmon.Probe.config_key == value)
        if self.editing:
            probe = self.get_instance()
            query = query.filter(tempmon.Probe.uuid != probe.uuid)
        if query.count():
            raise colander.Invalid(node, "Config key must be unique")

    def render_client(self, probe, field):
        client = probe.client
        if not client:
            return ""
        text = six.text_type(client)
        url = self.request.route_url('tempmon.clients.view', uuid=client.uuid)
        return tags.link_to(text, url)

    def render_appliance(self, probe, field):
        appliance = probe.appliance
        if not appliance:
            return ""
        text = six.text_type(appliance)
        url = self.request.route_url('tempmon.appliances.view', uuid=appliance.uuid)
        return tags.link_to(text, url)

    def render_enabled_form(self, probe, field):
        if probe.enabled:
            return raw_datetime(self.rattail_config, probe.enabled)
        return "No"

    def delete_instance(self, probe):
        # bulk-delete all readings first
        readings = self.Session.query(tempmon.Reading)\
                               .filter(tempmon.Reading.probe == probe)
        readings.delete(synchronize_session=False)
        self.Session.flush()
        self.Session.refresh(probe)

        # Flush immediately to force any pending integrity errors etc.; that
        # way we don't set flash message until we know we have success.
        self.Session.delete(probe)
        self.Session.flush()

    def get_row_data(self, probe):
        query = self.Session.query(tempmon.Reading)\
                            .filter(tempmon.Reading.probe == probe)
        return query

    def get_parent(self, reading):
        return reading.client

    def configure_row_grid(self, g):
        super(TempmonProbeView, self).configure_row_grid(g)

        # # probe
        # g.set_filter('probe', tempmon.Probe.description)
        # g.set_sorter('probe', tempmon.Probe.description)

        g.set_sort_defaults('taken', 'desc')

    def graph(self):
        probe = self.get_instance()
        use_buefy = self.get_use_buefy()

        key = 'tempmon.probe.{}.graph_time_range'.format(probe.uuid)
        selected = self.request.params.get('time-range')
        if not selected:
            selected = self.request.session.get(key, 'last hour')
        self.request.session[key] = selected

        range_options = tags.Options([
            tags.Option("Last Hour", 'last hour'),
            tags.Option("Last 6 Hours", 'last 6 hours'),
            tags.Option("Last Day", 'last day'),
            tags.Option("Last Week", 'last week'),
        ])

        if use_buefy:
            time_range = HTML.tag('b-select', c=[range_options.render()],
                                  **{'v-model': 'currentTimeRange',
                                     '@input': 'timeRangeChanged'})
        else:
            time_range = tags.select('time-range', selected, range_options)

        context = {
            'probe': probe,
            'parent_title': six.text_type(probe),
            'parent_url': self.get_action_url('view', probe),
            'time_range': time_range,
            'current_time_range': selected,
        }
        return self.render_to_response('graph', context)

    def graph_readings(self):
        probe = self.get_instance()

        key = 'tempmon.probe.{}.graph_time_range'.format(probe.uuid)
        selected = self.request.params['time-range']
        assert selected
        self.request.session[key] = selected

        # figure out what our window of time is
        if selected == 'last hour':
            cutoff = 60 * 60            # seconds x minutes
        elif selected == 'last 6 hours':
            cutoff = 60 * 60 * 6        # hour x 6
        elif selected == 'last day':
            cutoff = 60 * 60 * 24       # hour x 24
        elif selected == 'last week':
            cutoff = 60 * 60 * 24 * 7   # day x 7
        else:
            raise NotImplementedError("Unknown time range: {}".format(selected))

        # figure out which readings we need to graph
        cutoff = make_utc() - datetime.timedelta(seconds=cutoff)
        readings = self.Session.query(tempmon.Reading)\
                               .filter(tempmon.Reading.probe == probe)\
                               .filter(tempmon.Reading.taken >= cutoff)\
                               .order_by(tempmon.Reading.taken)\
                               .all()

        # convert readings to data for scatter plot
        data = [{
            'x': localtime(self.rattail_config, reading.taken, from_utc=True).isoformat(),
            'y': float(reading.degrees_f),
        } for reading in readings]
        return data

    @classmethod
    def defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        model_key = cls.get_model_key()
        permission_prefix = cls.get_permission_prefix()
        model_title_plural = cls.get_model_title_plural()

        # graph
        config.add_route('{}.graph'.format(route_prefix), '{}/{{{}}}/graph'.format(url_prefix, model_key),
                         request_method='GET')
        config.add_view(cls, attr='graph', route_name='{}.graph'.format(route_prefix),
                        permission='{}.view'.format(permission_prefix))

        # graph_readings
        config.add_route('{}.graph_readings'.format(route_prefix), '{}/{{{}}}/graph-readings'.format(url_prefix, model_key),
                         request_method='GET')
        config.add_view(cls, attr='graph_readings', route_name='{}.graph_readings'.format(route_prefix),
                        permission='{}.view'.format(permission_prefix), renderer='json')

        cls._defaults(config)


def includeme(config):
    TempmonProbeView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Department Views
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.db import model

from deform import widget as dfwidget

from tailbone import grids
from tailbone.views import MasterView, AutocompleteView


class DepartmentsView(MasterView):
    """
    Master view for the Department class.
    """
    model_class = model.Department
    touchable = True
    has_versions = True

    grid_columns = [
        'number',
        'name',
        'product',
        'personnel',
        'exempt_from_gross_sales',
    ]

    form_fields = [
        'number',
        'name',
        'product',
        'personnel',
        'exempt_from_gross_sales',
    ]

    def configure_grid(self, g):
        super(DepartmentsView, self).configure_grid(g)
        g.filters['name'].default_active = True
        g.filters['name'].default_verb = 'contains'
        g.set_sort_defaults('number')
        g.set_type('product', 'boolean')
        g.set_type('personnel', 'boolean')
        g.set_link('number')
        g.set_link('name')

    def configure_form(self, f):
        super(DepartmentsView, self).configure_form(f)
        f.remove_field('subdepartments')
        f.remove_field('employees')
        f.set_type('product', 'boolean')
        f.set_type('personnel', 'boolean')

    def template_kwargs_view(self, **kwargs):
        department = kwargs['instance']
        if department.employees:
            employees = sorted(department.employees, key=six.text_type)
            actions = [
                grids.GridAction('view', icon='zoomin',
                                 url=lambda r, i: self.request.route_url('employees.view', uuid=r.uuid))
            ]
            kwargs['employees'] = grids.Grid(None, employees, ['display_name'], request=self.request,
                                             model_class=model.Employee, main_actions=actions)
        else:
            kwargs['employees'] = None
        return kwargs

    def before_delete(self, department):
        """
        Check to see if there are any products which belong to the department;
        if there are then we do not allow delete and redirect the user.
        """
        count = self.Session.query(model.Product)\
                            .filter(model.Product.department == department)\
                            .count()
        if count:
            self.request.session.flash("Will not delete department which still has {} products: {}".format(
                count, department), 'error')
            raise self.redirect(self.get_action_url('view', department))

    def list_by_vendor(self):
        """
        View list of departments by vendor
        """
        data = self.Session.query(model.Department)\
                           .outerjoin(model.Product)\
                           .join(model.ProductCost)\
                           .join(model.Vendor)\
                           .filter(model.Vendor.uuid == self.request.params['uuid'])\
                           .distinct()\
                           .order_by(model.Department.name)

        def configure(g):
            g.configure(include=[
                g.name,
            ], readonly=True)

        def row_attrs(row, i):
            return {'data-uuid': row.uuid}

        grid = self.make_grid(data=data, sortable=False, filterable=False, pageable=False,
                              configure=configure, width=None, checkboxes=True,
                              row_attrs=row_attrs, main_actions=[], more_actions=[])
        self.request.response.content_type = str('text/html')
        self.request.response.text = grid.render_grid()
        return self.request.response

    @classmethod
    def defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()

        # list by vendor
        config.add_route('{}.by_vendor'.format(route_prefix), '{}/by-vendor'.format(url_prefix))
        config.add_view(cls, attr='list_by_vendor', route_name='{}.by_vendor'.format(route_prefix),
                        permission='{}.list'.format(permission_prefix))

        cls._defaults(config)


class DepartmentsAutocomplete(AutocompleteView):

    mapped_class = model.Department
    fieldname = 'name'


def includeme(config):

    # autocomplete
    config.add_route('departments.autocomplete',        '/departments/autocomplete')
    config.add_view(DepartmentsAutocomplete, route_name='departments.autocomplete',
                    renderer='json', permission='departments.list')

    DepartmentsView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - "Common" Views
"""

from __future__ import unicode_literals, absolute_import

import rattail
from rattail.db import model
from rattail.mail import send_email
from rattail.util import OrderedDict

from cornice import Service

import tailbone
from tailbone import forms
from tailbone.forms.common import Feedback
from tailbone.api import APIView, api
from tailbone.db import Session


class CommonView(APIView):
    """
    Misc. "common" views for the API.

    .. attribute:: feedback_email_key

       This is the email key which will be used when sending "user feedback"
       email.  Default value is ``'user_feedback'``.
    """
    feedback_email_key = 'user_feedback'

    @api
    def about(self):
        """
        Generic view to show "about project" info page.
        """
        packages = self.get_packages()
        return {
            'project_title': self.get_project_title(),
            'project_version': self.get_project_version(),
            'packages': packages,
            'package_names': list(packages),
        }

    def get_project_title(self):
        return self.rattail_config.app_title(default="Tailbone")

    def get_project_version(self):
        import tailbone
        return tailbone.__version__

    def get_packages(self):
        """
        Should return the full set of packages which should be displayed on the
        'about' page.
        """
        return OrderedDict([
            ('rattail', rattail.__version__),
            ('Tailbone', tailbone.__version__),
        ])

    @api
    def feedback(self):
        """
        View to handle user feedback form submits.
        """
        # TODO: this logic was copied from tailbone.views.common and is largely
        # identical; perhaps should merge somehow?
        schema = Feedback().bind(session=Session())
        form = forms.Form(schema=schema, request=self.request)
        if form.validate(newstyle=True):
            data = dict(form.validated)

            # figure out who the sending user is, if any
            if self.request.user:
                data['user'] = self.request.user
            elif data['user']:
                data['user'] = Session.query(model.User).get(data['user'])

            # TODO: should provide URL to view user
            if data['user']:
                data['user_url'] = '#' # TODO: could get from config?

            data['client_ip'] = self.request.client_addr
            send_email(self.rattail_config, self.feedback_email_key, data=data)
            return {'ok': True}

        return {'error': "Form did not validate!"}

    @classmethod
    def defaults(cls, config):

        # about
        about = Service(name='about', path='/about')
        about.add_view('GET', 'about', klass=cls)
        config.add_cornice_service(about)

        # feedback
        feedback = Service(name='feedback', path='/feedback')
        feedback.add_view('POST', 'feedback', klass=cls,
                          permission='common.feedback')
        config.add_cornice_service(feedback)


def includeme(config):
    CommonView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Template Context Helpers
"""

from __future__ import unicode_literals, absolute_import

import os
import datetime
from decimal import Decimal

from rattail.time import localtime, make_utc
from rattail.util import (pretty_quantity, pretty_hours, hours_as_decimal,
                          OrderedDict)
from rattail.db.util import maxlen

from webhelpers2.html import *
from webhelpers2.html.tags import *

from tailbone.util import csrf_token, get_csrf_token, pretty_datetime, raw_datetime


def pretty_date(date):
    """
    Render a human-friendly date string.
    """
    if not date:
        return ''
    return date.strftime('%a %d %b %Y')


def render_attrs(**attrs):
    """
    Convenience wrapper to replace the deprecated
    `webhelpers.html.builder.format_attrs()`
    """
    HTML.optimize_attrs(attrs)
    return HTML.render_attrs(attrs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - Inventory Batches
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail import pod
from rattail.db import model
from rattail.util import pretty_quantity

from cornice import Service

from tailbone.api.batch import APIBatchView, APIBatchRowView


class InventoryBatchViews(APIBatchView):

    model_class = model.InventoryBatch
    default_handler_spec = 'rattail.batch.inventory:InventoryBatchHandler'
    route_prefix = 'inventory'
    permission_prefix = 'batch.inventory'
    collection_url_prefix = '/inventory-batches'
    object_url_prefix = '/inventory-batch'
    supports_toggle_complete = True

    def normalize(self, batch):
        data = super(InventoryBatchViews, self).normalize(batch)

        data['mode'] = batch.mode
        data['mode_display'] = self.enum.INVENTORY_MODE.get(batch.mode)
        if data['mode_display'] is None and batch.mode is not None:
            data['mode_display'] = six.text_type(batch.mode)

        data['reason_code'] = batch.reason_code

        return data

    def count_modes(self):
        """
        Retrieve info about the available batch count modes.
        """
        permission_prefix = self.get_permission_prefix()
        if self.request.is_root:
            modes = self.handler.get_count_modes()
        else:
            modes = self.handler.get_allowed_count_modes(
                self.Session(), self.request.user,
                permission_prefix=permission_prefix)
        return modes

    def adjustment_reasons(self):
        """
        Retrieve info about the available "reasons" for inventory adjustment
        batches.
        """
        raw_reasons = self.handler.get_adjustment_reasons(self.Session())
        reasons = []
        for reason in raw_reasons:
            reasons.append({
                'uuid': reason.uuid,
                'code': reason.code,
                'description': reason.description,
                'hidden': reason.hidden,
            })
        return reasons

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)
        cls._batch_defaults(config)
        cls._inventory_defaults(config)

    @classmethod
    def _inventory_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        collection_url_prefix = cls.get_collection_url_prefix()

        # get count modes
        count_modes = Service(name='{}.count_modes'.format(route_prefix),
                              path='{}/count-modes'.format(collection_url_prefix))
        count_modes.add_view('GET', 'count_modes', klass=cls,
                             permission='{}.list'.format(permission_prefix))
        config.add_cornice_service(count_modes)

        # get adjustment reasons
        adjustment_reasons = Service(name='{}.adjustment_reasons'.format(route_prefix),
                                     path='{}/adjustment-reasons'.format(collection_url_prefix))
        adjustment_reasons.add_view('GET', 'adjustment_reasons', klass=cls,
                                    permission='{}.list'.format(permission_prefix))
        config.add_cornice_service(adjustment_reasons)


class InventoryBatchRowViews(APIBatchRowView):

    model_class = model.InventoryBatchRow
    default_handler_spec = 'rattail.batch.inventory:InventoryBatchHandler'
    route_prefix = 'inventory.rows'
    permission_prefix = 'batch.inventory'
    collection_url_prefix = '/inventory-batch-rows'
    object_url_prefix = '/inventory-batch-row'
    editable = True
    supports_quick_entry = True

    def normalize(self, row):
        batch = row.batch
        data = super(InventoryBatchRowViews, self).normalize(row)

        data['item_id'] = row.item_id
        data['upc'] = six.text_type(row.upc)
        data['upc_pretty'] = row.upc.pretty() if row.upc else None
        data['brand_name'] = row.brand_name
        data['description'] = row.description
        data['size'] = row.size
        data['full_description'] = row.product.full_description if row.product else row.description
        data['image_url'] = pod.get_image_url(self.rattail_config, row.upc) if row.upc else None
        data['case_quantity'] = pretty_quantity(row.case_quantity or 1)

        data['cases'] = row.cases
        data['units'] = row.units
        data['unit_uom'] = 'LB' if row.product and row.product.weighed else 'EA'
        data['quantity_display'] = "{} {}".format(
            pretty_quantity(row.cases or row.units),
            'CS' if row.cases else data['unit_uom'])

        data['allow_cases'] = self.handler.allow_cases(batch)

        return data

    def update_object(self, row, data):
        """
        Supplements the default logic as follows:

        Converts certain fields within the data, to proper "native" types.
        """
        # convert some data types as needed
        if 'cases' in data:
            if data['cases'] == '':
                data['cases'] = None
            elif data['cases']:
                data['cases'] = int(data['cases'])
        if 'units' in data:
            if data['units'] == '':
                data['units'] = None
            elif data['units']:
                data['units'] = int(data['units'])

        # update row per usual
        row = super(InventoryBatchRowViews, self).update_object(row, data)
        return row


def includeme(config):
    InventoryBatchViews.defaults(config)
    InventoryBatchRowViews.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Setup script for Tailbone
"""

from __future__ import unicode_literals, absolute_import

import os.path
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))
exec(open(os.path.join(here, 'tailbone', '_version.py')).read())
README = open(os.path.join(here, 'README.rst')).read()


requires = [
    #
    # Version numbers within comments below have specific meanings.
    # Basically the 'low' value is a "soft low," and 'high' a "soft high."
    # In other words:
    #
    # If either a 'low' or 'high' value exists, the primary point to be
    # made about the value is that it represents the most current (stable)
    # version available for the package (assuming typical public access
    # methods) whenever this project was started and/or documented.
    # Therefore:
    #
    # If a 'low' version is present, you should know that attempts to use
    # versions of the package significantly older than the 'low' version
    # may not yield happy results.  (A "hard" high limit may or may not be
    # indicated by a true version requirement.)
    #
    # Similarly, if a 'high' version is present, and especially if this
    # project has laid dormant for a while, you may need to refactor a bit
    # when attempting to support a more recent version of the package.  (A
    # "hard" low limit should be indicated by a true version requirement
    # when a 'high' version is present.)
    #
    # In any case, developers and other users are encouraged to play
    # outside the lines with regard to these soft limits.  If bugs are
    # encountered then they should be filed as such.
    #
    # package                           # low                   high

    # TODO: why do we need to cap this?  breaks tailbone.db zope stuff somehow
    'zope.sqlalchemy<1.0',              # 0.7                   0.7.7

    # TODO: apparently they jumped from 0.1 to 0.9 and that broke us...
    # (0.1 was released on 2014-09-14 and then 0.9 came out on 2018-09-27)
    # (i've cached 0.1 at pypi.rattailproject.org just in case it disappears)
    # (still, probably a better idea is to refactor so we can use 0.9)
    'webhelpers2_grid==0.1',            # 0.1

    # TODO: remove version cap once we can drop support for python 2.x
    'cornice<5.0',                      # 3.4.2                 4.0.1

    # TODO: remove once their bug is fixed?  idk what this is about yet...
    'deform<2.0.15',                    # 2.0.14

    'colander',                         # 1.7.0
    'ColanderAlchemy',                  # 0.3.3
    'humanize',                         # 0.5.1
    'Mako',                             # 0.6.2
    'openpyxl',                         # 2.4.7
    'paginate',                         # 0.5.6
    'paginate_sqlalchemy',              # 0.2.0
    'passlib',                          # 1.7.1
    'Pillow',                           # 5.3.0
    'pyramid',                          # 1.3b2
    'pyramid_beaker>=0.6',              #                       0.6.1
    'pyramid_deform',                   # 0.2
    'pyramid_exclog',                   # 0.6
    'pyramid_mako',                     # 1.0.2
    'pyramid_tm',                       # 0.3
    'rattail[db,bouncer]',              # 0.5.0
    'six',                              # 1.10.0
    'sqlalchemy-filters',               # 0.8.0
    'transaction',                      # 1.2.0
    'waitress',                         # 0.8.1
    'WebHelpers2',                      # 2.0
    'WTForms',                          # 2.1
]


extras = {

    'docs': [
        #
        # package                       # low                   high

        'Sphinx',                       # 1.2
        'sphinx-rtd-theme',             # 0.2.4
    ],

    'tests': [
        #
        # package                       # low                   high

        'coverage',                     # 3.6
        'fixture',                      # 1.5
        'mock',                         # 1.0.1
        'nose',                         # 1.3.0
    ],
}


setup(
    name = "Tailbone",
    version = __version__,
    author = "Lance Edgar",
    author_email = "lance@edbob.org",
    url = "http://rattailproject.org/",
    license = "GNU GPL v3",
    description = "Backoffice Web Application for Rattail",
    long_description = README,

    classifiers = [
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Office/Business',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],

    install_requires = requires,
    extras_require = extras,
    tests_require = ['Tailbone[tests]'],
    test_suite = 'nose.collector',

    packages = find_packages(exclude=['tests.*', 'tests']),
    include_package_data = True,
    zip_safe = False,

    entry_points = {

        'paste.app_factory': [
            'main = tailbone.app:main',
        ],

        'rattail.config.extensions': [
            'tailbone = tailbone.config:ConfigExtension',
        ],

        'pyramid.scaffold': [
            'rattail = tailbone.scaffolds:RattailTemplate',
        ],
    },
)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Common Forms
"""

from __future__ import unicode_literals, absolute_import

from rattail.db import model

import colander


@colander.deferred
def validate_user(node, kw):
    session = kw['session']
    def validate(node, value):
        user = session.query(model.User).get(value)
        if not user:
            raise colander.Invalid(node, "User not found")
        return user.uuid
    return validate


class Feedback(colander.Schema):
    """
    Form schema for user feedback.
    """
    referrer = colander.SchemaNode(colander.String())

    user = colander.SchemaNode(colander.String(),
                               missing=colander.null,
                               validator=validate_user)

    user_name = colander.SchemaNode(colander.String(),
                                    missing=colander.null)

    message = colander.SchemaNode(colander.String())



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Views for 'ordering' (purchasing) batches
"""

from __future__ import unicode_literals, absolute_import

import os
import json

import six
import openpyxl
from sqlalchemy import orm

from rattail.db import model, api
from rattail.core import Object
from rattail.time import localtime

from webhelpers2.html import tags

from tailbone.views.purchasing import PurchasingBatchView


class OrderingBatchView(PurchasingBatchView):
    """
    Master view for "ordering" batches.
    """
    route_prefix = 'ordering'
    url_prefix = '/ordering'
    model_title = "Ordering Batch"
    model_title_plural = "Ordering Batches"
    index_title = "Ordering"
    mobile_creatable = True
    rows_editable = True
    mobile_rows_creatable = True
    mobile_rows_quickable = True
    mobile_rows_editable = True
    mobile_rows_deletable = True
    has_worksheet = True

    labels = {
        'po_total_calculated': "PO Total",
    }

    form_fields = [
        'id',
        'store',
        'buyer',
        'vendor',
        'department',
        'purchase',
        'vendor_email',
        'vendor_fax',
        'vendor_contact',
        'vendor_phone',
        'date_ordered',
        'po_number',
        'po_total_calculated',
        'notes',
        'created',
        'created_by',
        'status_code',
        'complete',
        'executed',
        'executed_by',
    ]

    mobile_form_fields = [
        'vendor',
        'department',
        'date_ordered',
        'po_number',
        'po_total',
        'created',
        'created_by',
        'notes',
        'status_code',
        'complete',
        'executed',
        'executed_by',
    ]

    row_labels = {
        'po_total_calculated': "PO Total",
    }

    row_grid_columns = [
        'sequence',
        'upc',
        # 'item_id',
        'brand_name',
        'description',
        'size',
        'cases_ordered',
        'units_ordered',
        # 'cases_received',
        # 'units_received',
        'po_total_calculated',
        # 'invoice_total',
        # 'credits',
        'status_code',
    ]

    row_form_fields = [
        'item_entry',
        'item_id',
        'upc',
        'product',
        'brand_name',
        'description',
        'size',
        'case_quantity',
        'cases_ordered',
        'units_ordered',
        'po_line_number',
        'po_unit_cost',
        'po_total_calculated',
        'status_code',
    ]

    order_form_header_columns = [
        "UPC",
        "Brand",
        "Description",
        "Case",
        "Vend. Code",
        "Pref.",
        "Unit Cost",
    ]

    @property
    def batch_mode(self):
        return self.enum.PURCHASE_BATCH_MODE_ORDERING

    def configure_form(self, f):
        super(OrderingBatchView, self).configure_form(f)

        # purchase
        f.remove_field('purchase')

    def get_batch_kwargs(self, batch, mobile=False):
        kwargs = super(OrderingBatchView, self).get_batch_kwargs(batch, mobile=mobile)
        kwargs['ship_method'] = batch.ship_method
        kwargs['notes_to_vendor'] = batch.notes_to_vendor
        return kwargs

    def configure_row_form(self, f):
        """
        Supplements the default logic as follows:

        When editing, only these fields allow changes; all others are made
        read-only:

        * ``cases_ordered``
        * ``units_ordered``
        """
        super(OrderingBatchView, self).configure_row_form(f)

        # when editing, only certain fields should allow changes
        if self.editing:
            editable_fields = [
                'cases_ordered',
                'units_ordered',
            ]
            for field in f.fields:
                if field not in editable_fields:
                    f.set_readonly(field)

    def worksheet(self):
        """
        View for editing batch row data as an order form worksheet.
        """
        batch = self.get_instance()
        if batch.executed:
            return self.redirect(self.get_action_url('view', batch))

        # organize existing batch rows by product
        order_items = {}
        for row in batch.active_rows():
            order_items[row.product_uuid] = row

        # organize vendor catalog costs by dept / subdept
        departments = {}
        costs = self.handler.get_order_form_costs(self.Session(), batch.vendor)
        costs = self.handler.sort_order_form_costs(costs)
        costs = list(costs)   # we must have a stable list for the rest of this
        self.handler.decorate_order_form_costs(batch, costs)
        for cost in costs:

            department = cost.product.department
            if department:
                departments.setdefault(department.uuid, department)
            else:
                if None not in departments:
                    department = Object(name='', number=None)
                    departments[None] = department
                department = departments[None]
            
            subdepartments = getattr(department, '_order_subdepartments', None)
            if subdepartments is None:
                subdepartments = department._order_subdepartments = {}

            subdepartment = cost.product.subdepartment
            if subdepartment:
                subdepartments.setdefault(subdepartment.uuid, subdepartment)
            else:
                if None not in subdepartments:
                    subdepartment = Object(name=None, number=None)
                    subdepartments[None] = subdepartment
                subdepartment = subdepartments[None]

            subdept_costs = getattr(subdepartment, '_order_costs', None)
            if subdept_costs is None:
                subdept_costs = subdepartment._order_costs = []
            subdept_costs.append(cost)
            cost._batchrow = order_items.get(cost.product_uuid)

        # fetch recent purchase history, sort/pad for template convenience
        history = self.handler.get_order_form_history(batch, costs, 6)
        for i in range(6 - len(history)):
            history.append(None)
        history = list(reversed(history))

        title = self.get_instance_title(batch)
        order_date = batch.date_ordered
        if not order_date:
            order_date = localtime(self.rattail_config).date()

        buefy_data = None
        if self.get_use_buefy():
            buefy_data = self.get_worksheet_buefy_data(departments)

        return self.render_to_response('worksheet', {
            'batch': batch,
            'order_date': order_date,
            'instance': batch,
            'instance_title': title,
            'instance_url': self.get_action_url('view', batch),
            'vendor': batch.vendor,
            'departments': departments,
            'history': history,
            'get_upc': lambda p: p.upc.pretty() if p.upc else '',
            'header_columns': self.order_form_header_columns,
            'ignore_cases': not self.handler.allow_cases(),
            'worksheet_data': buefy_data,
        })

    def get_worksheet_buefy_data(self, departments):
        data = {}
        for department in six.itervalues(departments):
            for subdepartment in six.itervalues(department._order_subdepartments):
                for i, cost in enumerate(subdepartment._order_costs, 1):
                    cases = int(cost._batchrow.cases_ordered or 0) if cost._batchrow else None
                    units = int(cost._batchrow.units_ordered or 0) if cost._batchrow else None
                    key = 'cost_{}'.format(cost.uuid)
                    data['{}_cases'.format(key)] = cases
                    data['{}_units'.format(key)] = units

                    total = 0
                    row = cost._batchrow
                    if row:
                        total = row.po_total_calculated or row.po_total or 0
                    if not (total or cases or units):
                        display = ''
                    else:
                        display = '${:0,.2f}'.format(total)
                    data['{}_total_display'.format(key)] = display

        return data

    def worksheet_update(self):
        """
        Handles AJAX requests to update the order quantities for some row
        within the current batch, from the worksheet view.  POST data should
        include:

        * ``product_uuid``
        * ``cases_ordered``
        * ``units_ordered``

        If a row already exists for the given product, it will be updated;
        otherwise a new row is created for the product and then that is
        updated.  The handler's
        :meth:`~rattail:rattail.batch.purchase.PurchaseBatchHandler.update_row_quantity()`
        method is invoked to update the row.

        However, if both of the quantities given are empty, and a row exists
        for the given product, then that row is removed from the batch, instead
        of being updated.  If a matching row is not found, it will not be
        created.
        """
        batch = self.get_instance()

        try:
            data = self.request.json_body
        except json.JSONDecodeError:
            data = self.request.POST

        cases_ordered = data.get('cases_ordered')
        if cases_ordered is None:
            cases_ordered = 0
        elif not isinstance(cases_ordered, int):
            if cases_ordered == '':
                cases_ordered = 0
            else:
                cases_ordered = int(cases_ordered)
        if cases_ordered >= 100000: # TODO: really this depends on underlying column
            return {'error': "Invalid value for cases ordered: {}".format(cases_ordered)}

        units_ordered = data.get('units_ordered')
        if units_ordered is None:
            units_ordered = 0
        elif not isinstance(units_ordered, int):
            if units_ordered == '':
                units_ordered = 0
            else:
                units_ordered = int(units_ordered)
        if units_ordered >= 100000: # TODO: really this depends on underlying column
            return {'error': "Invalid value for units ordered: {}".format(units_ordered)}

        uuid = data.get('product_uuid')
        product = self.Session.query(model.Product).get(uuid) if uuid else None
        if not product:
            return {'error': "Product not found"}

        # first we find out which existing row(s) match the given product
        matches = [row for row in batch.active_rows()
                   if row.product_uuid == product.uuid]
        if matches and len(matches) != 1:
            raise RuntimeError("found too many ({}) matches for product {} in batch {}".format(
                len(matches), product.uuid, batch.uuid))

        row = None
        if cases_ordered or units_ordered:

            # make a new row if necessary
            if matches:
                row = matches[0]
            else:
                row = self.handler.make_row()
                row.product = product
                self.handler.add_row(batch, row)

            # update row quantities
            self.handler.update_row_quantity(row, cases_ordered=cases_ordered,
                                             units_ordered=units_ordered)

        else: # empty order quantities

            # remove row if present
            if matches:
                row = matches[0]
                self.handler.do_remove_row(row)
                row = None

        return {
            'row_cases_ordered': int(row.cases_ordered or 0) if row else None,
            'row_units_ordered': int(row.units_ordered or 0) if row else None,
            'row_po_total': '${:0,.2f}'.format(row.po_total or 0) if row else None,
            'row_po_total_calculated': '${:0,.2f}'.format(row.po_total_calculated or 0) if row else None,
            'row_po_total_display': '${:0,.2f}'.format(row.po_total_calculated or row.po_total or 0) if row else None,
            'batch_po_total': '${:0,.2f}'.format(batch.po_total or 0),
            'batch_po_total_calculated': '${:0,.2f}'.format(batch.po_total_calculated or 0),
            'batch_po_total_display': '${:0,.2f}'.format(batch.po_total_calculated or batch.po_total or 0),
        }

    def render_mobile_listitem(self, batch, i):
        return "({}) {} on {} for ${:0,.2f}".format(batch.id_str, batch.vendor,
                                                    batch.date_ordered, batch.po_total or 0)

    def mobile_create(self):
        """
        Mobile view for creating a new ordering batch
        """
        mode = self.batch_mode
        data = {'mode': mode}

        vendor = None
        if self.request.method == 'POST' and self.request.POST.get('vendor'):
            vendor = self.Session.query(model.Vendor).get(self.request.POST['vendor'])
            if vendor:

                # fetch first to avoid flush below
                store = self.rattail_config.get_store(self.Session())

                batch = self.model_class()
                batch.mode = mode
                batch.vendor = vendor
                batch.store = store
                batch.buyer = self.request.user.employee
                batch.created_by = self.request.user
                batch.po_total = 0
                kwargs = self.get_batch_kwargs(batch, mobile=True)
                batch = self.handler.make_batch(self.Session(), **kwargs)
                if self.handler.should_populate(batch):
                    self.handler.populate(batch)
                return self.redirect(self.request.route_url('mobile.ordering.view', uuid=batch.uuid))

        data['index_title'] = self.get_index_title()
        data['index_url'] = self.get_index_url(mobile=True)
        data['mode_title'] = self.enum.PURCHASE_BATCH_MODE[mode].capitalize()

        data['vendor_use_autocomplete'] = self.rattail_config.getbool(
            'rattail', 'vendor.use_autocomplete', default=True)
        if not data['vendor_use_autocomplete']:
            vendors = self.Session.query(model.Vendor)\
                                  .order_by(model.Vendor.name)
            options = [(tags.Option(vendor.name, vendor.uuid))
                       for vendor in vendors]
            options.insert(0, tags.Option("(please choose)", ''))
            data['vendor_options'] = options

        return self.render_to_response('create', data, mobile=True)

    def configure_mobile_row_form(self, f):
        super(OrderingBatchView, self).configure_mobile_row_form(f)
        if self.editing:
            # TODO: probably should take `allow_cases` into account here...
            f.focus_spec = '[name="units_ordered"]'

    def download_excel(self):
        """
        Download ordering batch as Excel spreadsheet.
        """
        batch = self.get_instance()

        # populate Excel worksheet
        workbook = openpyxl.Workbook()
        worksheet = workbook.active
        worksheet.title = "Purchase Order"
        worksheet.append(["Store", "Vendor", "Date ordered"])
        worksheet.append([batch.store.name, batch.vendor.name, batch.date_ordered.strftime('%m/%d/%Y')])
        worksheet.append([])
        worksheet.append(['vendor_code', 'upc', 'brand_name', 'description', 'cases_ordered', 'units_ordered'])
        for row in batch.active_rows():
            worksheet.append([row.vendor_code, six.text_type(row.upc), row.brand_name,
                              '{} {}'.format(row.description, row.size),
                              row.cases_ordered, row.units_ordered])

        # write Excel file to batch data dir
        filedir = batch.filedir(self.rattail_config)
        if not os.path.exists(filedir):
            os.makedirs(filedir)
        filename = 'PO.{}.xlsx'.format(batch.id_str)
        path = batch.filepath(self.rattail_config, filename)
        workbook.save(path)

        return self.file_response(path)

    @classmethod
    def _ordering_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        url_prefix = cls.get_url_prefix()
        permission_prefix = cls.get_permission_prefix()
        model_title = cls.get_model_title()
        model_title_plural = cls.get_model_title_plural()

        # fix permission group label
        config.add_tailbone_permission_group(permission_prefix, model_title_plural)

        # download as Excel
        config.add_route('{}.download_excel'.format(route_prefix), '{}/{{uuid}}/excel'.format(url_prefix))
        config.add_view(cls, attr='download_excel', route_name='{}.download_excel'.format(route_prefix),
                        permission='{}.download_excel'.format(permission_prefix))
        config.add_tailbone_permission(permission_prefix, '{}.download_excel'.format(permission_prefix),
                                       "Download {} as Excel".format(model_title))

    @classmethod
    def defaults(cls, config):
        cls._ordering_defaults(config)
        cls._purchasing_defaults(config)
        cls._batch_defaults(config)
        cls._defaults(config)


def includeme(config):
    OrderingBatchView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - Auth Views
"""

from __future__ import unicode_literals, absolute_import

from rattail.db.auth import authenticate_user, set_user_password, cache_permissions

from cornice import Service

from tailbone.api import APIView, api
from tailbone.db import Session
from tailbone.auth import login_user, logout_user


class AuthenticationView(APIView):

    @api
    def check_session(self):
        """
        View to serve as "no-op" / ping action to check current user's session.
        This will establish a server-side web session for the user if none
        exists.  Note that this also resets the user's session timer.
        """
        data = {'ok': True}
        if self.request.user:
            data['user'] = self.get_user_info(self.request.user)

        data['permissions'] = list(self.request.tailbone_cached_permissions)

        # background color may be set per-request, by some apps
        if hasattr(self.request, 'background_color') and self.request.background_color:
            data['background_color'] = self.request.background_color
        else: # otherwise we use the one from config
            data['background_color'] = self.rattail_config.get(
                'tailbone', 'background_color')

        return data

    @api
    def login(self):
        """
        API login view.
        """
        if self.request.method == 'OPTIONS':
            return self.request.response

        username = self.request.json.get('username')
        password = self.request.json.get('password')
        if not (username and password):
            return {'error': "Invalid username or password"}

        # make sure credentials are valid
        user = self.authenticate_user(username, password)
        if not user:
            return {'error': "Invalid username or password"}

        # is there some reason this user should not login?
        error = self.why_cant_user_login(user)
        if error:
            return {'error': error}

        login_user(self.request, user)
        return {
            'ok': True,
            'user': self.get_user_info(user),
            'permissions': list(cache_permissions(Session(), user)),
        }

    def authenticate_user(self, username, password):
        return authenticate_user(Session(), username, password)

    def why_cant_user_login(self, user):
        """
        This method is given a ``User`` instance, which represents someone who
        is just now trying to login, and has already cleared the basic hurdle
        of providing the correct credentials for a user on file.  This method
        is responsible then, for further verification that this user *should*
        in fact be allowed to login to this app node.  If the method determines
        a reason the user should *not* be allowed to login, then it should
        return that reason as a simple string.
        """

    @api
    def logout(self):
        """
        API logout view.
        """
        if self.request.method == 'OPTIONS':
            return self.request.response

        logout_user(self.request)
        return {'ok': True}

    @api
    def become_root(self):
        """
        Elevate the current request to 'root' for full system access.
        """
        if not self.request.is_admin:
            raise self.forbidden()
        self.request.user.record_event(self.enum.USER_EVENT_BECOME_ROOT)
        self.request.session['is_root'] = True
        return {
            'ok': True,
            'user': self.get_user_info(self.request.user),
        }

    @api
    def stop_root(self):
        """
        Lower the current request from 'root' back to normal access.
        """
        if not self.request.is_admin:
            raise self.forbidden()
        self.request.user.record_event(self.enum.USER_EVENT_STOP_ROOT)
        self.request.session['is_root'] = False
        return {
            'ok': True,
            'user': self.get_user_info(self.request.user),
        }

    @api
    def change_password(self):
        """
        View which allows a user to change their password.
        """
        if self.request.method == 'OPTIONS':
            return self.request.response

        if not self.request.user:
            raise self.forbidden()

        data = self.request.json_body

        # first make sure "current" password is accurate
        if not authenticate_user(Session(), self.request.user, data['current_password']):
            return {'error': "The current/old password you provided is incorrect"}

        # okay then, set new password
        set_user_password(self.request.user, data['new_password'])
        return {
            'ok': True,
            'user': self.get_user_info(self.request.user),
        }

    @classmethod
    def defaults(cls, config):
        cls._auth_defaults(config)

    @classmethod
    def _auth_defaults(cls, config):

        # session
        check_session = Service(name='check_session', path='/session')
        check_session.add_view('GET', 'check_session', klass=cls)
        config.add_cornice_service(check_session)

        # login
        login = Service(name='login', path='/login')
        login.add_view('POST', 'login', klass=cls)
        config.add_cornice_service(login)

        # logout
        logout = Service(name='logout', path='/logout')
        logout.add_view('POST', 'logout', klass=cls)
        config.add_cornice_service(logout)

        # become root
        become_root = Service(name='become_root', path='/become-root')
        become_root.add_view('POST', 'become_root', klass=cls)
        config.add_cornice_service(become_root)

        # stop root
        stop_root = Service(name='stop_root', path='/stop-root')
        stop_root.add_view('POST', 'stop_root', klass=cls)
        config.add_cornice_service(stop_root)

        # change password
        change_password = Service(name='change_password', path='/change-password')
        change_password.add_view('POST', 'change_password', klass=cls)
        config.add_cornice_service(change_password)


def includeme(config):
    AuthenticationView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Tailbone Web API - Batch Views
"""

from __future__ import unicode_literals, absolute_import

import logging

import six

from rattail.time import localtime
from rattail.util import load_object

from cornice import resource, Service

from tailbone.api import APIMasterView2 as APIMasterView


log = logging.getLogger(__name__)


class APIBatchMixin(object):
    """
    Base class for all API views which are meant to handle "batch" *and/or*
    "batch row" data.
    """

    def get_batch_class(self):
        model_class = self.get_model_class()
        if hasattr(model_class, '__batch_class__'):
            return model_class.__batch_class__
        return model_class

    def get_handler(self):
        """
        Returns a `BatchHandler` instance for the view.  All (?) custom batch
        API views should define a default handler class; however this may in all
        (?)  cases be overridden by config also.  The specific setting required
        to do so will depend on the 'key' for the type of batch involved, e.g.
        assuming the 'vendor_catalog' batch:

        .. code-block:: ini

           [rattail.batch]
           vendor_catalog.handler = myapp.batch.vendorcatalog:CustomCatalogHandler

        Note that the 'key' for a batch is generally the same as its primary
        table name, although technically it is whatever value returns from the
        ``batch_key`` attribute of the main batch model class.
        """
        key = self.get_batch_class().batch_key
        spec = self.rattail_config.get('rattail.batch', '{}.handler'.format(key),
                                       default=self.default_handler_spec)
        return load_object(spec)(self.rattail_config)


class APIBatchView(APIBatchMixin, APIMasterView):
    """
    Base class for all API views which are meant to handle "batch" *and/or*
    "batch row" data.
    """
    supports_toggle_complete = False
    supports_execute = False

    def __init__(self, request, **kwargs):
        super(APIBatchView, self).__init__(request, **kwargs)
        self.handler = self.get_handler()

    def normalize(self, batch):

        created = localtime(self.rattail_config, batch.created, from_utc=True)

        executed = None
        if batch.executed:
            executed = localtime(self.rattail_config, batch.executed, from_utc=True)

        return {
            'uuid': batch.uuid,
            '_str': six.text_type(batch),
            'id': batch.id,
            'id_str': batch.id_str,
            'description': batch.description,
            'notes': batch.notes,
            'params': batch.params or {},
            'rowcount': batch.rowcount,
            'created': six.text_type(created),
            'created_display': self.pretty_datetime(created),
            'created_by_uuid': batch.created_by.uuid,
            'created_by_display': six.text_type(batch.created_by),
            'complete': batch.complete,
            'status_code': batch.status_code,
            'status_display': batch.STATUS.get(batch.status_code,
                                               six.text_type(batch.status_code)),
            'executed': six.text_type(executed) if executed else None,
            'executed_display': self.pretty_datetime(executed) if executed else None,
            'executed_by_uuid': batch.executed_by_uuid,
            'executed_by_display': six.text_type(batch.executed_by or ''),
        }

    def create_object(self, data):
        """
        Create a new object instance and populate it with the given data.

        Here we'll invoke the handler for actual batch creation, instead of
        typical logic used for simple records.
        """
        user = self.request.user
        kwargs = dict(data)
        kwargs['user'] = user
        batch = self.handler.make_batch(self.Session(), **kwargs)
        if self.handler.should_populate(batch):
            self.handler.do_populate(batch, user)
        return batch

    def update_object(self, batch, data):
        """
        Logic for updating a main object record.

        Here we want to make sure we set "created by" to the current user, when
        creating a new batch.
        """
        # we're only concerned with *new* batches here
        if not batch.uuid:

            # assign creator; initialize row count
            batch.created_by_uuid = self.request.user.uuid
            if batch.rowcount is None:
                batch.rowcount = 0

        # then go ahead with usual logic
        return super(APIBatchView, self).update_object(batch, data)

    def mark_complete(self):
        """
        Mark the given batch as "complete".
        """
        batch = self.get_object()

        if batch.executed:
            return {'error': "Batch {} has already been executed: {}".format(
                batch.id_str, batch.description)}

        if batch.complete:
            return {'error': "Batch {} is already marked complete: {}".format(
                batch.id_str, batch.description)}

        batch.complete = True
        return self._get(obj=batch)

    def mark_incomplete(self):
        """
        Mark the given batch as "incomplete".
        """
        batch = self.get_object()

        if batch.executed:
            return {'error': "Batch {} has already been executed: {}".format(
                batch.id_str, batch.description)}

        if not batch.complete:
            return {'error': "Batch {} is already marked incomplete: {}".format(
                batch.id_str, batch.description)}

        batch.complete = False
        return self._get(obj=batch)

    def execute(self):
        """
        Execute the given batch.
        """
        batch = self.get_object()

        if batch.executed:
            return {'error': "Batch {} has already been executed: {}".format(
                batch.id_str, batch.description)}

        kwargs = dict(self.request.json_body)
        kwargs.pop('user', None)
        kwargs.pop('progress', None)
        result = self.handler.do_execute(batch, self.request.user, **kwargs)
        return {'ok': bool(result), 'batch': self.normalize(batch)}

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)
        cls._batch_defaults(config)

    @classmethod
    def _batch_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        collection_url_prefix = cls.get_collection_url_prefix()
        object_url_prefix = cls.get_object_url_prefix()

        if cls.supports_toggle_complete:

            # mark complete
            mark_complete = Service(name='{}.mark_complete'.format(route_prefix),
                                    path='{}/{{uuid}}/mark-complete'.format(object_url_prefix))
            mark_complete.add_view('POST', 'mark_complete', klass=cls,
                                   permission='{}.edit'.format(permission_prefix))
            config.add_cornice_service(mark_complete)

            # mark incomplete
            mark_incomplete = Service(name='{}.mark_incomplete'.format(route_prefix),
                                      path='{}/{{uuid}}/mark-incomplete'.format(object_url_prefix))
            mark_incomplete.add_view('POST', 'mark_incomplete', klass=cls,
                                     permission='{}.edit'.format(permission_prefix))
            config.add_cornice_service(mark_incomplete)

        if cls.supports_execute:

            # execute batch
            execute = Service(name='{}.execute'.format(route_prefix),
                              path='{}/{{uuid}}/execute'.format(object_url_prefix))
            execute.add_view('POST', 'execute', klass=cls,
                             permission='{}.execute'.format(permission_prefix))
            config.add_cornice_service(execute)


# TODO: deprecate / remove this
BatchAPIMasterView = APIBatchView


class APIBatchRowView(APIBatchMixin, APIMasterView):
    """
    Base class for all API views which are meant to handle "batch rows" data.
    """
    editable = False
    supports_quick_entry = False

    def __init__(self, request, **kwargs):
        super(APIBatchRowView, self).__init__(request, **kwargs)
        self.handler = self.get_handler()

    def normalize(self, row):
        batch = row.batch
        return {
            'uuid': row.uuid,
            '_str': six.text_type(row),
            '_parent_str': six.text_type(batch),
            '_parent_uuid': batch.uuid,
            'batch_uuid': batch.uuid,
            'batch_id': batch.id,
            'batch_id_str': batch.id_str,
            'batch_description': batch.description,
            'batch_complete': batch.complete,
            'batch_executed': bool(batch.executed),
            'sequence': row.sequence,
            'status_code': row.status_code,
            'status_display': row.STATUS.get(row.status_code, six.text_type(row.status_code)),
        }

    def update_object(self, row, data):
        """
        Supplements the default logic as follows:

        Invokes the batch handler's ``refresh_row()`` method after updating the
        row's field data per usual.
        """
        # update row per usual
        row = super(APIBatchRowView, self).update_object(row, data)

        # okay now we apply handler refresh logic
        self.handler.refresh_row(row)
        return row

    def delete_object(self, row):
        """
        Overrides the default logic as follows:

        Delegates deletion of the row to the batch handler.
        """
        self.handler.do_remove_row(row)

    def quick_entry(self):
        """
        View for handling "quick entry" user input, for a batch.
        """
        data = self.request.json_body

        uuid = data['batch_uuid']
        batch = self.Session.query(self.get_batch_class()).get(uuid)
        if not batch:
            raise self.notfound()

        entry = data['quick_entry']

        try:
            row = self.handler.quick_entry(self.Session(), batch, entry)
        except Exception as error:
            log.warning("quick entry failed for '%s' batch %s: %s",
                        self.handler.batch_key, batch.id_str, entry,
                        exc_info=True)
            msg = six.text_type(error)
            if not msg and isinstance(error, NotImplementedError):
                msg = "Feature is not implemented"
            return {'error': msg}

        self.Session.flush()
        result = self._get(obj=row)
        result['ok'] = True
        return result

    @classmethod
    def defaults(cls, config):
        cls._defaults(config)
        cls._batch_row_defaults(config)

    @classmethod
    def _batch_row_defaults(cls, config):
        route_prefix = cls.get_route_prefix()
        permission_prefix = cls.get_permission_prefix()
        collection_url_prefix = cls.get_collection_url_prefix()
        object_url_prefix = cls.get_object_url_prefix()

        if cls.supports_quick_entry:

            # quick entry
            config.add_route('{}.quick_entry'.format(route_prefix), '{}/quick-entry'.format(collection_url_prefix),
                             request_method=('OPTIONS', 'POST'))
            config.add_view(cls, attr='quick_entry', route_name='{}.quick_entry'.format(route_prefix),
                            permission='{}.edit'.format(permission_prefix),
                            renderer='json')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Customer Order Views
"""

from __future__ import unicode_literals, absolute_import

import six
from sqlalchemy import orm

from rattail.db import model

from webhelpers2.html import tags

from tailbone.db import Session
from tailbone.views import MasterView


class CustomerOrdersView(MasterView):
    """
    Master view for customer orders
    """
    model_class = model.CustomerOrder
    route_prefix = 'custorders'
    editable = False
    deletable = False

    grid_columns = [
        'id',
        'customer',
        'person',
        'created',
        'status_code',
    ]

    form_fields = [
        'id',
        'customer',
        'person',
        'phone_number',
        'email_address',
        'created',
        'status_code',
    ]

    def query(self, session):
        return session.query(model.CustomerOrder)\
                      .options(orm.joinedload(model.CustomerOrder.customer))

    def configure_grid(self, g):
        super(CustomerOrdersView, self).configure_grid(g)

        g.set_joiner('customer', lambda q: q.outerjoin(model.Customer))
        g.set_joiner('person', lambda q: q.outerjoin(model.Person))

        g.filters['customer'] = g.make_filter('customer', model.Customer.name,
                                              label="Customer Name",
                                              default_active=True,
                                              default_verb='contains')
        g.filters['person'] = g.make_filter('person', model.Person.display_name,
                                            label="Person Name",
                                            default_active=True,
                                            default_verb='contains')

        g.set_sorter('customer', model.Customer.name)
        g.set_sorter('person', model.Person.display_name)

        g.set_sort_defaults('created', 'desc')

        # TODO: enum choices renderer
        g.set_label('status_code', "Status")
        g.set_label('id', "ID")

    def configure_form(self, f):
        super(CustomerOrdersView, self).configure_form(f)

        # id
        f.set_readonly('id')
        f.set_label('id', "ID")

        # person
        f.set_renderer('person', self.render_person)

        # created
        f.set_readonly('created')

        # label overrides
        f.set_label('status_code', "Status")

    def render_person(self, order, field):
        person = order.person
        if not person:
            return ""
        text = six.text_type(person)
        url = self.request.route_url('people.view', uuid=person.uuid)
        return tags.link_to(text, url)

    def create(self, form=None, template='create'):
        """
        View for creating a new customer order.  Note that it does so by way of
        maintaining a "new customer order" batch, until the user finally
        submits the order, at which point the batch is converted to a proper
        order.
        """
        batch = self.get_current_batch()

        if self.request.method == 'POST':

            # first we check for traditional form post
            action = self.request.POST.get('action')
            post_actions = [
                'start_over_entirely',
                'delete_batch',
            ]
            if action in post_actions:
                return getattr(self, action)(batch)

            # okay then, we'll assume newer JSON-style post params
            data = dict(self.request.json_body)
            action = data.get('action')
            json_actions = [
                'get_customer_info',
                'set_customer_data',
                'submit_new_order',
            ]
            if action in json_actions:
                result = getattr(self, action)(batch, data)
                return self.json_response(result)

        context = {'batch': batch}
        return self.render_to_response(template, context)

    def get_current_batch(self):
        user = self.request.user
        if not user:
            raise RuntimeError("this feature requires a user to be logged in")

        try:
            # there should be at most *one* new batch per user
            batch = self.Session.query(model.CustomerOrderBatch)\
                                .filter(model.CustomerOrderBatch.mode == self.enum.CUSTORDER_BATCH_MODE_CREATING)\
                                .filter(model.CustomerOrderBatch.created_by == user)\
                                .one()

        except orm.exc.NoResultFound:
            # no batch yet for this user, so make one
            batch = model.CustomerOrderBatch()
            batch.mode = self.enum.CUSTORDER_BATCH_MODE_CREATING
            batch.created_by = user
            self.Session.add(batch)
            self.Session.flush()

        return batch

    def start_over_entirely(self, batch):
        # just delete current batch outright
        # TODO: should use self.handler.do_delete() instead?
        self.Session.delete(batch)
        self.Session.flush()

        # send user back to normal "create" page; a new batch will be generated
        # for them automatically
        route_prefix = self.get_route_prefix()
        url = self.request.route_url('{}.create'.format(route_prefix))
        return self.redirect(url)

    def delete_batch(self, batch):
        # just delete current batch outright
        # TODO: should use self.handler.do_delete() instead?
        self.Session.delete(batch)
        self.Session.flush()

        # set flash msg just to be more obvious
        self.request.session.flash("New customer order has been deleted.")

        # send user back to customer orders page, w/ no new batch generated
        route_prefix = self.get_route_prefix()
        url = self.request.route_url(route_prefix)
        return self.redirect(url)

    def get_customer_info(self, batch, data):
        uuid = data.get('uuid')
        if not uuid:
            return {'error': "Must specify a customer UUID"}

        customer = self.Session.query(model.Customer).get(uuid)
        if not customer:
            return {'error': "Customer not found"}

        return self.info_for_customer(batch, data, customer)

    def info_for_customer(self, batch, data, customer):
        phone = customer.first_phone()
        email = customer.first_email()
        return {
            'uuid': customer.uuid,
            'phone_number': phone.number if phone else None,
            'email_address': email.address if email else None,
        }

    def set_customer_data(self, batch, data):
        if 'customer_uuid' in data:
            batch.customer_uuid = data['customer_uuid']
            if 'person_uuid' in data:
                batch.person_uuid = data['person_uuid']
            elif batch.customer_uuid:
                self.Session.flush()
                batch.person = batch.customer.first_person()
            else: # no customer set
                batch.person_uuid = None
        if 'phone_number' in data:
            batch.phone_number = data['phone_number']
        if 'email_address' in data:
            batch.email_address = data['email_address']
        self.Session.flush()
        return {'success': True}

    def submit_new_order(self, batch, data):
        # TODO
        return {'success': True}


def includeme(config):
    CustomerOrdersView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
FileMon Views
"""

from __future__ import unicode_literals, absolute_import

import subprocess
import logging

from tailbone.views import View


log = logging.getLogger(__name__)


class FilemonView(View):
    """
    Misc. views for Filemon...(for now)
    """

    def restart(self):
        cmd = self.rattail_config.getlist('tailbone', 'filemon.restart', default='/bin/sleep 3') # simulate by default
        log.debug("attempting filemon restart with command: %s", cmd)
        try:
            subprocess.check_call(cmd)
        except Except as error:
            self.request.session.flash("FileMon daemon could not be restarted: {}".format(error), 'error')
        else:
            self.request.session.flash("FileMon daemon has been restarted.")
        return self.redirect(self.request.get_referrer(default=self.request.route_url('datasyncchanges')))

    @classmethod
    def defaults(cls, config):

        # fix permission group title
        config.add_tailbone_permission_group('filemon', label="FileMon")

        # restart filemon
        config.add_tailbone_permission('filemon', 'filemon.restart', label="Restart FileMon Daemon")
        config.add_route('filemon.restart', '/filemon/restart', request_method='POST')
        config.add_view(cls, attr='restart', route_name='filemon.restart', permission='filemon.restart')


def includeme(config):
    FilemonView.defaults(config)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base class for customer order batch views
"""

from __future__ import unicode_literals, absolute_import

import six

from rattail.db import model

import colander

from tailbone import forms
from tailbone.views.batch import BatchMasterView


class CustomerOrderBatchView(BatchMasterView):
    """
    Master view base class, for customer order batches.  The views for the
    various mode/workflow batches will derive from this.
    """
    model_class = model.CustomerOrderBatch
    model_row_class = model.CustomerOrderBatchRow
    default_handler_spec = 'rattail.batch.custorder:CustomerOrderBatchHandler'

    grid_columns = [
        'id',
        'customer',
        'rows',
        'created',
        'created_by',
    ]

    form_fields = [
        'id',
        'customer',
        'person',
        'phone_number',
        'email_address',
        'created',
        'created_by',
        'rows',
        'status_code',
    ]

    def configure_grid(self, g):
        super(CustomerOrderBatchView, self).configure_grid(g)

        g.set_link('customer')
        g.set_link('created')
        g.set_link('created_by')

    def configure_form(self, f):
        super(CustomerOrderBatchView, self).configure_form(f)
        order = f.model_instance
        model = self.rattail_config.get_model()

        # readonly fields
        f.set_readonly('rows')
        f.set_readonly('status_code')

        # customer
        if 'customer' in f.fields and self.editing:
            f.replace('customer', 'customer_uuid')
            f.set_node('customer_uuid', colander.String(), missing=colander.null)
            customer_display = ""
            if self.request.method == 'POST':
                if self.request.POST.get('customer_uuid'):
                    customer = self.Session.query(model.Customer)\
                                           .get(self.request.POST['customer_uuid'])
                    if customer:
                        customer_display = six.text_type(customer)
            elif self.editing:
                customer_display = six.text_type(order.customer or "")
            customers_url = self.request.route_url('customers.autocomplete')
            f.set_widget('customer_uuid', forms.widgets.JQueryAutocompleteWidget(
                field_display=customer_display, service_url=customers_url))
            f.set_label('customer_uuid', "Customer")
        else:
            f.set_renderer('customer', self.render_customer)

        # person
        if 'person' in f.fields and self.editing:
            f.replace('person', 'person_uuid')
            f.set_node('person_uuid', colander.String(), missing=colander.null)
            person_display = ""
            if self.request.method == 'POST':
                if self.request.POST.get('person_uuid'):
                    person = self.Session.query(model.Person)\
                                         .get(self.request.POST['person_uuid'])
                    if person:
                        person_display = six.text_type(person)
            elif self.editing:
                person_display = six.text_type(order.person or "")
            people_url = self.request.route_url('people.autocomplete')
            f.set_widget('person_uuid', forms.widgets.JQueryAutocompleteWidget(
                field_display=person_display, service_url=people_url))
            f.set_label('person_uuid', "Person")
        else:
            f.set_renderer('person', self.render_person)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Form Schema Types
"""

from __future__ import unicode_literals, absolute_import

import re
import datetime

import six

from rattail.db import model
from rattail.gpc import GPC

import colander

from tailbone.db import Session


class JQueryTime(colander.Time):
    """
    Custom type for jQuery widget Time data.
    """

    def deserialize(self, node, cstruct):
        if not cstruct:
            return colander.null

        formats = [
            '%I:%M %p',
            '%I:%M%p',
            '%I %p',
            '%I%p',
        ]
        for fmt in formats:
            try:
                return datetime.datetime.strptime(cstruct, fmt).time()
            except ValueError:
                pass

        # re-try first format, for "better" error message
        return datetime.datetime.strptime(cstruct, formats[0]).time()


class DateTimeBoolean(colander.Boolean):
    """
    Schema type which presents the user with a "boolean" whereas the underlying
    node is really a datetime (assumed to be "naive" UTC, and allow nulls).
    """

    def deserialize(self, node, cstruct):
        value = super(DateTimeBoolean, self).deserialize(node, cstruct)
        if value: # else return None
            return datetime.datetime.utcnow()


class GPCType(colander.SchemaType):
    """
    Schema type for product GPC data.
    """

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null
        return six.text_type(appstruct)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return None
        digits = re.sub(r'\D', '', cstruct)
        if not digits:
            return None
        try:
            return GPC(digits)
        except Exception as err:
            raise colander.Invalid(node, six.text_type(err))


class ProductQuantity(colander.MappingSchema):
    """
    Combo schema type for product cases and units; useful for inventory,
    ordering, receiving etc.  Meant to be used with the ``CasesUnitsWidget``.
    """
    cases = colander.SchemaNode(colander.Decimal(), missing=colander.null)

    units = colander.SchemaNode(colander.Decimal(), missing=colander.null)


class ModelType(colander.SchemaType):
    """
    Custom schema type for scalar ORM relationship fields.
    """
    model_class = None
    session = None

    def __init__(self, model_class=None, session=None):
        if model_class:
            self.model_class = model_class
        if session:
            self.session = session
        else:
            self.session = self.make_session()

    def make_session(self):
        return Session()

    @property
    def model_title(self):
        self.model_class.get_model_title()

    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null
        return six.text_type(appstruct)

    def deserialize(self, node, cstruct):
        if not cstruct:
            return None
        obj = self.session.query(self.model_class).get(cstruct)
        if not obj:
            raise colander.Invalid(node, "{} not found".format(self.model_title))
        return obj


# TODO: deprecate / remove this
ObjectType = ModelType


class StoreType(ModelType):
    """
    Custom schema type for store field.
    """
    model_class = model.Store


class CustomerType(ModelType):
    """
    Custom schema type for customer field.
    """
    model_class = model.Customer


class DepartmentType(ModelType):
    """
    Custom schema type for department field.
    """
    model_class = model.Department


class EmployeeType(ModelType):
    """
    Custom schema type for employee field.
    """
    model_class = model.Employee


class VendorType(ModelType):
    """
    Custom schema type for vendor relationship field.
    """
    model_class = model.Vendor


class ProductType(ModelType):
    """
    Custom schema type for product relationship field.
    """
    model_class = model.Product


class UserType(ModelType):
    """
    Custom schema type for user field.
    """
    model_class = model.User
